
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include "dl_list.h"

struct dlst_entry_s {
	dlst_entry_t next, prev;
	void *data;
};

struct dlst_list_s {
	dlst_entry_t head, tail;
	size_t size;
};

void dlst_init(dlst_t *l) {
	*l = malloc(sizeof(struct dlst_list_s));
	(*l)->size = 0;
	(*l)->head = (*l)->tail = NULL;
}

void dlst_free(dlst_t *l) {
	dlst_clear(*l);
	free(*l);
}

void dlst_free_data(dlst_t *l, void (*f)(void *)) {
	dlst_clear_data(*l, f);
	free(*l);
}

dlst_entry_t dlst_push(dlst_t l, void *d) {
	dlst_entry_t e = malloc(sizeof(struct dlst_entry_s));
	e->next = NULL;
	e->prev = l->tail;
	e->data = d;
	if (l->size++ > 0) {
		l->tail->next = e;
		l->tail = e;
	} else {
		l->head = l->tail = e;
	}
	return e;
}

void * dlst_pop(dlst_t l) {
	dlst_entry_t e = l->head;
	void *r = e->data;
	if (--l->size) {
		l->head = e->next;
		l->head->prev = NULL;
	} else {
		l->head = l->tail = NULL;
	}
	free(e);
	return r;
}

size_t dlst_size(const dlst_t l) {
	return l->size;
}

void dlst_clear_data(dlst_t l, void (*f)(void *)) {
	dlst_entry_t e = l->head;
	l->head = NULL;
	l->tail = NULL;
	l->size = 0;
	if (f) {
		while (e) {
			dlst_entry_t en = e->next;
			f(e->data);
			free(e);
			e = en;
		}
	} else {
		while (e) {
			dlst_entry_t en = e->next;
			free(en);
			e = en;
		}
	}
}

void dlst_clear(dlst_t l) {
	dlst_entry_t e = l->head;
	l->head = l->tail = NULL;
	l->size = 0;
	while (e) {
		dlst_entry_t en = e->next;
		free(e);
		e = en;
	}
}

dlst_entry_t dlst_first(const dlst_t l) {
	return l->head;
}

dlst_entry_t dlst_last(const dlst_t l) {
	return l->tail;
}

dlst_entry_t dlst_next(const dlst_t l, const dlst_entry_t e) {
	return e->next;
}

dlst_entry_t dlst_prev(const dlst_t l, const dlst_entry_t e) {
	return e->prev;
}

bool dlst_empty(dlst_t l) {
	return !(l->size);
}

void * dlst_data(const dlst_t l, const dlst_entry_t e) {
	return e->data;
}

void * dlst_remove(dlst_t l, const dlst_entry_t e) {
	if (e->prev == NULL) {
		l->head = e->next;
	} else {
		e->prev->next = e->next;
	}
	
	if (e->next == NULL) {
		l->tail = e->prev;
	} else {
		e->next->prev = e->prev;
	}
	
	l->size--;
	void *r = e->data;
	free(e);
	return r;
}

dlst_entry_t dlst_insert(dlst_t l, const dlst_entry_t o, void * d, const bool after) {
	dlst_entry_t e = malloc(sizeof(struct dlst_entry_s));
	e->data = d;
	
	if (!l->size) {
		e->prev = e->next = NULL;
		return l->head = l->tail = e;
	}
	
	if (after) {
		e->prev = o;
		e->next = o->next;
		if (o == l->tail) {
			l->tail = e;
		} else {
			o->next->prev = e;
		}
		o->next = e;
	} else {
		e->next = o;
		e->prev = o->prev;
		if (o == l->head) {
			l->head = e;
		} else {
			o->prev->next = e;
		}
		o->prev = e;
	}
	return e;
}
/*
int main() {
	list l;
	lst_init(&l);
	
	
	printf("insert: %08x\n", lst_push(l, (void *)6));
	printf("insert: %08x\n", lst_push(l, (void *)7));
 	printf("%d\n", lst_pop(l));
	printf("insert: %08x\n", lst_push(l, (void *)8));
	
	printf("head %08x, tail %08x\n", l->head, l->tail);
	entry e;
	lst_forall(e, l)
 		printf("%08x, data: %d\n", e, lst_remove(l, e));
 	
 	while (lst_size(l))
 		printf("%d\n", lst_pop(l));
 	
	lst_free(&l);
	return 0;
}*/
