
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdint.h>		/* for uint64_t */
#include <limits.h>		/* for ULLONG_MAX */
#include <signal.h>		/* for signal() */

#include "dl_list.h"

#define _DECL_OPTS
#define _DECL_INTERRUPTED
#include "fdf.h"

volatile bool interrupted = false;
struct option_s opts = { false, false, false, NORMAL, NULL };

extern int errno;

#define PRINT_BUF_SIZE 256
static char print_buf[PRINT_BUF_SIZE];

char *format_size(off_t s) {
	static const char unit[] = { 'b', 'k', 'm', 'g', 't', 'p', 'e', 'z', 'y' };
	unsigned int i;
	long double d = s;
	for (i = 0; i < sizeof(unit) / sizeof(unit[0]) && d >= 10240.0; i++)
		d /= 1024.0;
	if (i) {
		off_t t = (off_t)d;
		if ((long double)t == d) {
			snprintf(print_buf, PRINT_BUF_SIZE, "%llu %c", t, unit[i]);
		} else {
			snprintf(print_buf, PRINT_BUF_SIZE, "%.1Lf %c", d, unit[i]);
		}
	} else {
		snprintf(print_buf, PRINT_BUF_SIZE, "%llu %c", s, unit[i]);
	}
	return print_buf;
}

static char * path_concat(const char *a, const size_t a_len, const char *b, const size_t b_len) {
	const bool needs_slash = (a[a_len - 1] != '/');
	size_t off = 0;
	char *dest_path = malloc(a_len + needs_slash + b_len + 1);
	if (!dest_path) {
		perror(_("error allocating memory"));
		return NULL;
	}
	strncpy(dest_path, a, a_len);
	off += a_len;
	if (needs_slash)
		dest_path[off++] = '/';
	strncpy(dest_path + off, b, b_len);
	off += b_len;
	dest_path[off] = '\0';
	return dest_path;
}

static inline fentry_s * new_fentry_s(char *path, flags_t flags, range_t **const range, unsigned int input_nr) {
	fentry_s *fe = malloc(sizeof(fentry_s));
	fe->path = path;
	fe->flags = flags;
	fe->range = range;
	fe->input_nr = input_nr;
	return fe;
}

static inline fentry_s * new_fentry_t(char *path, const fentry_s *fe) {
	return new_fentry_s(path, fe->flags, fe->range, fe->input_nr);
}

static bool check_dir(dlst_t srcs, dlst_entry_t e, const fentry_s *fe) {
	struct dirent **dirs;
	
	int n = scandir(fe->path, &dirs, 0, alphasort);
	if (n < 0) {
		ERROR(_("error scanning directory '%s': %s\n"), fe->path, strerror(errno));
		return false;
	}
	
	const char *src = fe->path;
	const size_t src_len = strlen(src);
	
	DBG("scandir returned %d\n", n);
	
	while (n--) {
		const char *d_name = dirs[n]->d_name;
		const size_t len = strlen(d_name);
		switch (len) {
			case 1:
			case 2:	// break when dirs[n] is '.' or '..'
				if (!strncmp("..", d_name, len))
					break;
				// fall-through otherwise
			default: {
				DBG(_("in dir: d_name '%s', src '%s'\n"), d_name, fe->path);
			//	dlst_insert_after(srcs, e, fe);
				dlst_push(srcs, new_fentry_t(path_concat(src, src_len, d_name, len), fe));
			} break;
		}
		free(dirs[n]);
	}
	free(dirs);
	return true;
}

char * readlink_malloc(char *path, ssize_t *length) {
	ssize_t buflen = 100;
	char *buf = malloc(buflen);
	ssize_t len;
	while (true) {
		len = readlink(path, buf, buflen);
		if (len == -1)
			break;
		if (len < buflen) {
			buf[len] = '\0';
			if (length)
				*length = len;
			return buf;
		}
		buflen *= 2;
		if (buflen <= 0) {
			errno = ENAMETOOLONG;
			break;
		}
		buf = realloc(buf, buflen);
	}
	free(buf);
	return NULL;
}

static bool check_sources(dlst_t srcs, dlst_entry_t e) {
	struct stat s;
	bool device_set = false;
	dev_t device = 0;
	dlst_entry_t en;
	
	while (!interrupted && e && !dlst_empty(srcs)) {
		fentry_s *fe = dlst_data(srcs, e);
		if (lstat(fe->path, &s)) {
			ERROR(_("error accessing file '%s': %s\n"), fe->path, strerror(errno));
			goto err;
		}
		
		const char *type = NULL;
		const char *expl = NULL;
		
		if (!device_set) {
			device = s.st_dev;
			device_set = true;
		} else if (fe->flags.stay_on_fs && device != s.st_dev) {
			type = "";
			expl = _("not on same device as initial file");
			goto err_info;
		} else if (fe->range) {
			range_t *r = *fe->range;
			if (s.st_size < r->lower || s.st_size > r->upper) {
				char *fsize = strdup(format_size(s.st_size));
				snprintf(print_buf, PRINT_BUF_SIZE, _("file-size %s is not in range %s"), fsize, r->spec);
				free(fsize);
				type = "";
				expl = print_buf;
				goto err_info;
			}
		}
		
		fe->blksize = s.st_blksize;
		fe->size = s.st_size;
		fe->mode = s.st_mode;
		fe->gid = s.st_gid;
		fe->uid = s.st_uid;
		
		switch (s.st_mode & S_IFMT) {
			case S_IFSOCK:
				type = _("socket ");
				goto err_info;
			case S_IFLNK:
				if (fe->flags.follow_symlinks) {
					char *ln_path = readlink_malloc(fe->path, NULL);		// TODO: don't need readlink_malloc
					if (ln_path) {
						dlst_push(srcs, new_fentry_t(ln_path, fe));
					} else {
						ERROR(_("error reading symbolic link '%s': %s\n"), fe->path, strerror(errno));
						goto err;
					}
				}
				if (!fe->flags.scan_symlinks) {
					type = _("symbolic link ");
					goto err_info;
				}
				break;
			case S_IFREG: // regular file
				break;
			case S_IFBLK:
				type = _("block device ");
				goto err_info;
			case S_IFDIR:
				if (fe->flags.single) {
					type = _("directory ");
					goto err_info;
				}
				check_dir(srcs, e, fe);
				goto err;
			case S_IFCHR:
				type = _("character device ");
				goto err_info;
			case S_IFIFO:
				type = _("named pipe ");
				goto err_info;
			default:
				type = _("unknown file-type ");
				snprintf(print_buf, PRINT_BUF_SIZE, _("mode: %o"), s.st_mode);
				expl = print_buf;
				goto err_info;
		}
	
		e = dlst_next(srcs, e);
		continue;
		
err_info:
		if (expl) {
			INFO(_("won't scan %s'%s': %s\n"), type, fe->path, expl);
		} else {
			INFO(_("won't scan %s'%s'\n"), type, fe->path);
		}
err:
		en = dlst_next(srcs, e);
		dlst_remove(srcs, e);
		e = en;
	}
	return true;
}

static bool read_stdin(dlst_t srcs, flags_t flags, range_t **range, unsigned int input_nr, const bool zeroterm) {
	ssize_t read;
	char *line = NULL;
	size_t len = 0;
	while ((read = getline(&line, &len, stdin)) != -1) {
		if (zeroterm) {
			char *tline = line;
			size_t tlen = len, end;
			do {
				end = min(strlen(tline), tlen);
				if (end == 0 || read < 0)
					break;
				dlst_entry_t e = dlst_push(srcs, new_fentry_s(strndup(tline, tlen), flags, range, input_nr));
				check_sources(srcs, e);
				tline += end;
			} while (zeroterm && (tlen -= end) > 0);
		} else {
			size_t tlen = read;
			if (line[tlen - 1] == '\n')
				line[--tlen] = '\0';
			dlst_entry_t e = dlst_push(srcs, new_fentry_s(strndup(line, tlen), flags, range, input_nr));
			check_sources(srcs, e);
		}
	}
	if (line)
		free(line);
	if (!feof(stdin)) {
		perror("getline()");
		return false;
	}
	return true;
}

static void fentry_free(void *v) {
	fentry_s *fe = v;
	free(fe->path);
	if (fe->range && *fe->range) {
		free((*fe->range)->spec);
		free(*fe->range);
		*fe->range = NULL;
	}
	free(fe);
}

static int comp_fentry(const void *v, const void *w) {
	off_t ov = (*((fentry_s **)v))->size;
	off_t ow = (*((fentry_s **)w))->size;
	return (ov < ow) ? -1 : (ov > ow) ? 1 : 0;
}

static int comp_fname(const void *v, const void *w) {
	fentry_s *fev = *(fentry_s **)v;
	fentry_s *few = *(fentry_s **)w;
	int r = strcmp(basename(fev->path), basename(few->path));
	return signum(r);
}

static void for_unique_range(
		fentry_s **entries, unsigned int from, size_t to, int (*comp)(const void *, const void *),
		void (*exec)(fentry_s **, unsigned int, unsigned int)) {
	unsigned int i = from;
	while (i < to && !interrupted) {
		unsigned int ss = i;
		int r = 0;
		while (r == 0 && ++i < to)
			r = comp(&(entries[ss]), &(entries[i]));
		
		if (i - ss > 1)
			exec(entries, ss, i);
	}
}

static void check_range(fentry_s **entries, unsigned int from, unsigned int to) {
	MSG(_("scan(entries, %d, %d)\n"), from, to);
	
	unsigned int i;
	dlst_t l;
	dlst_init(&l);
	fentry_s *fe = NULL;
	for (i=from; i<to; i++) {
		fe = entries[i];
		dlst_push(l, fe);
		MSG(_("\t%s (%llu bytes, idx: %u)\n"), fe->path, fe->size, i);
	}
	check_list(l, fe->size);
	
	dlst_free(&l);
}

static void check_range_fnames(fentry_s **entries, unsigned int from, unsigned int to) {
	DBG("qsorting .. from: %d, to: %d\n", from, to);
	qsort(&(entries[from]), to - from, sizeof(fentry_s *), comp_fname);
	for_unique_range(entries, from, to, comp_fname, check_range);
}

static bool check_files(fentry_s **entries, size_t len) {
	INFO(_("found %u files...\n"), len);
	qsort(entries, len, sizeof(fentry_s *), comp_fentry);
	
	if (opts.msglevel == DEBUG) {
		unsigned int i;
		for (i=0; i<len; i++) {
			fentry_s *fe = entries[i];
			DBG("%llu\t%s\n", fe->size, fe->path);
		}
	}
	
	for_unique_range(entries, 0, len, comp_fentry, (opts.fname_match) ? check_range_fnames : check_range);
	bool r = true;
	return r;
}

static bool output_started = false;
static unsigned int output_nr;

void start_output() {
	if (!output_started) {
		output_started = true;
		INFO("Input#\tPref\tChunk#\tPath\n");
	}
	output_nr = 0;
}

void output_fe(const fentry_s *fe, bool pref) {
	printf("%d\t%c\t%d\t%s\n",
			fe->input_nr,
			(pref) ? 'P' : (fe->flags.preferred) ? 'p' : '.',
			output_nr++,
			fe->path);
}

void stop_output() {
	printf("\n");
}

void handler(int sig) {
	switch (sig) {
		case SIGINT:
		case SIGTERM:
			interrupted = true;
	}
	signal(sig, handler);
}

static void usage(char *);
static range_t *parse_range(char *);

extern char *optarg;
extern int optind, opterr, optopt;

int main(int argc, char **argv) {
	char *program_name = *argv;
	
	int r;
	int opt;
	FILE *out_file = stdout;
	
	signal(SIGINT, handler);
	signal(SIGTERM, handler);
	
	dlst_t srcs;
	dlst_init(&srcs);
	
	bool inzeroterm = false;
	union {
		unsigned char val;
		flags_t flags;
	} f = { 0 };
	range_t *range = NULL;
	
	unsigned int input_nr = 1;
	
	while ((opt = getopt(argc, argv, "-rsxl:cdpo:0zfkvDqh")) != -1) {
		switch (opt) {
			case 1:
				if (!memcmp("-", optarg, sizeof("-"))) {
					bool ret = read_stdin(srcs, f.flags, (range) ? &range->p : NULL, input_nr, inzeroterm);
					if (!ret && opts.failfast)
						goto error;
				} else {
					fentry_s *fe = new_fentry_s(strdup(optarg), f.flags, (range) ? &range->p : NULL, input_nr);
					dlst_entry_t e = dlst_push(srcs, fe);
					if (!check_sources(srcs, e) && opts.failfast) goto error;
				}
				inzeroterm = false;
				f.val = 0;
				range = NULL;
				input_nr++;
				break;
			case 'c': f.flags.scan_symlinks = true; break;
			case 'r': f.flags.single = true; break;
			case 's': f.flags.follow_symlinks = true; break;
			case 'x': f.flags.stay_on_fs = true; break;
			case 'l':
				errno = 0;
				range = parse_range(optarg);
				if (errno) {
					ERROR(_("invalid range '%s'\n"), optarg);
					if (range != NULL)
						free(range);
					goto error;
				} else if (range != NULL && (uint64_t)range->upper < (uint64_t)range->lower) {
					ERROR(_("upper < lower for range '%s'\n"), optarg);
					free(range);
					goto error;
				}
				range->p = range;
				range->spec = strdup(optarg);
				break;
			case 'd': f.flags.dirout = true; break;
			case 'p': f.flags.preferred = true; break;
			case 'o': 
				out_file = fopen(optarg, "w+");
				if (!out_file) {
					perror(_("error opening output-file"));
					goto error;
				}
				break;
			case '0': inzeroterm = true; break;
			case 'z': opts.zeroterm = true; break;
			case 'f': opts.fname_match = true; break;
			case 'k': opts.failfast = true; break;
			case 'v': opts.msglevel = VERBOSE; break;
			case 'D': opts.msglevel = DEBUG; break;
			case 'q': opts.msglevel = QUIET; break;
			case 'h': usage(program_name); goto end;
			default:
				fprintf(stderr, _("Use '-h' to view a list of all supported options\n"));
				goto error;
		}
	}
	
	if (dlst_empty(srcs)) {
		fprintf(stderr, _("no files to check\n"));
		goto error;
	}
	
	size_t len = dlst_size(srcs);
	fentry_s **entries = malloc(len * sizeof(fentry_s *));
	dlst_entry_t e;
	unsigned int i = 0;
	dlst_forall(e, srcs)
		entries[i++] = dlst_data(srcs, e);
	
	if (!check_files(entries, len))
		goto error;
	
end:
	r = EXIT_SUCCESS;
	if (false)
error:	r = EXIT_FAILURE;
	
	if (out_file != stdout && fclose(out_file)) {
		perror(_("error closing output file"));
		r = EXIT_FAILURE;
	}
	dlst_free_data(&srcs, fentry_free);
	return r;
}

static range_t *parse_range(char *s) {
	if (*s == '\0')
		return NULL;
	
	off_t lower = 0;
	off_t upper = ~(off_t)0;
	char *endptr;
	unsigned long long int val;
	if (*s != '-') {
		val = strtoull(s, &endptr, 0);
		unsigned int m;
		switch (*endptr) {
			case '-': m = 0; break;
			case 'b': case 'B': m =  0; endptr++; break;
			case 'k': case 'K': m = 10; endptr++; break;
			case 'm': case 'M': m = 20; endptr++; break;
			case 'g': case 'G': m = 30; endptr++; break;
			case 't': case 'T': m = 40; endptr++; break;
			default:
				errno = EINVAL;
				return NULL;
		}
		lower = min(ULLONG_MAX >> m, val) << m;
		s = endptr;
	}
	
	if (*s == '-') do {
		s++;
		if (*s == '\0') break;
		if (*s == '-') {
			errno = EINVAL;
			return NULL;
		}
		val = strtoull(s, &endptr, 0);
		unsigned int m;
		switch (*endptr) {
			case '\0': m = 0; break;
			case 'b': case 'B': m =  0; endptr++; break;
			case 'k': case 'K': m = 10; endptr++; break;
			case 'm': case 'M': m = 20; endptr++; break;
			case 'g': case 'G': m = 30; endptr++; break;
			case 't': case 'T': m = 40; endptr++; break;
			default:
				errno = EINVAL;
				return NULL;
		}
		if (*endptr != '\0') {
			errno = EINVAL;
			return NULL;
		}
		upper = min(ULLONG_MAX >> m, val) << m;
	} while (false);
	
	if (lower == (off_t)0 && upper == ~(off_t)0)
		return NULL;
	range_t *r = malloc(sizeof(range_t));
	r->lower = lower;
	r->upper = upper;
	return r;
}

static void usage(char *program_name) {
	puts(_("Test for double files"));
	printf(_("Usage: %s [OPTS] { - | FILE } [ [OPTS] FILE ... ]\n"), program_name);
	puts("");
	puts(_("Options affecting only subsequently specified input:"));
	puts(_("  -r        don't recurse into directories"));
	puts(_("  -s        follow symbolic links (can't handle recusive links yet)"));
	puts(_("  -x        stay on the filesystem the input resides on (recursive mode)"));
	puts(_("  -l RANGE  limit files to scan to those whose size is in RANGE"));
	puts(_("  -c        scan content of symbolic links and identify doubles"));/*
	puts(_("\
  -d       if tested directories contain >= 1 file and this file is tested\n\
           positively, output directory instead of all the contained file(s)"));*/
	puts(_("\
  -p        mark input as preferred, output only doubles from this input when\n\
            matching files from other inputs are found"));
	puts("");
	puts(_("\
RANGE is specified by [SIZE]-[SIZE], whereas SIZE is a positive integer which\n\
may be post-fixed by 'b', 'k', 'm', 'g', 't', for bytes, kilo-, mega-, giga-\n\
bytes, etc. The lower or upper limit can be omitted."));
	puts("");
	puts(_("Other options:"));
	puts(_("  -o FILE   redirect output to FILE"));
	puts(_("  -0        separate file names from stdin by 0x00-bytes instead of newline"));
	puts(_("  -z        when writing file names, separate by 0x00-bytes instead of newline"));
	puts(_("  -f        file names of doubles must match as well"));
	puts(_("  -k        fail fast"));
	puts(_("  -v        verbose output"));
	puts(_("  -D        debug output (generally not wanted)"));
	puts(_("  -q        suppress any output (except for -o option)"));
	puts(_("  -h        shows this help message"));
	puts("");
	puts(_("\
This program is Free Software and released under the terms of the General Public\n\
License v2 by Franz Brausse (<dev@karlchenofhell.org>) in the hope to be useful."));
}
